import { Reader } from './reader.js';
import { Analyzer } from './analyzer.js';
import Mustache from 'mustache';
import * as fs from 'fs';
import * as del from 'del';
import moment from 'moment';
import { dirname } from 'path';
import { fileURLToPath } from 'url';
export class Reporter {
    constructor() {
        this.theme = 'white';
        this.folderToReadReport = `${process.cwd()}/`;
        this.folderToWriteReport = `${process.cwd()}/report/gherkin-detailer/`;
        this.folderToReadTemplates = `${dirname(fileURLToPath(import.meta.url))}/../templates/`;
        this.gherkins = [];
        this.templates = {};
        this.templatesView = {};
        this.templatePartials = {};
        this.reader = new Reader();
        this.analyzer = new Analyzer();
        this.reportFeaturesFiles = this.reportFeaturesFiles.bind(this);
    }
    createGherkinsReport(config) {
        this.theme = config.theme;
        this.folderToReadReport = config.analysisFolder || this.folderToReadReport;
        this.folderToWriteReport = config.outputFolder || this.folderToWriteReport;
        if (this.folderToWriteReport.substr(-1) !== '/') {
            this.folderToWriteReport += '/';
        }
        this.setupReportFolder();
        this.reader.readFeatureFilesFromFolder(this.folderToReadReport, this.reportFeaturesFiles);
    }
    setupReportFolder() {
        if (fs.existsSync(this.folderToWriteReport)) {
            del.deleteSync(this.folderToWriteReport);
        }
        fs.mkdirSync(this.folderToWriteReport, { recursive: true });
        fs.copyFileSync(`${this.folderToReadTemplates}${this.theme}.css`, `${this.folderToWriteReport}style.css`);
    }
    async readAllGherkins(readFiles) {
        let rowsFiles = [];
        await Promise.all(readFiles
            .map(async (readFile, indexFile) => {
            const contentFile = await this.reader.readContentFeatureFile(readFile);
            const rowsFile = this.reader.getRowsFeatureFile(contentFile);
            const gherkinsFile = this.analyzer.getGherkins(rowsFile, indexFile);
            const dataFile = { ...gherkinsFile, file: readFile.replace(process.cwd(), '.') };
            rowsFiles.push(dataFile);
        }));
        let gherkinsTexts = rowsFiles.map((rowFile) => {
            return rowFile.files.filter((file) => file.id !== '').map((file) => file.text);
        });
        gherkinsTexts = [].concat(...gherkinsTexts);
        let gherkinsObjects = rowsFiles.map((rowFile) => {
            return rowFile.files.filter((file) => file.id !== '').map((file) => file);
        });
        gherkinsObjects = [].concat(...gherkinsObjects);
        const similarities = this.analyzer.getSimilarities(gherkinsTexts, gherkinsObjects);
        rowsFiles = rowsFiles.map((rowFile) => {
            rowFile.files = rowFile.files.map((file) => {
                return this.addSimilarityInfo(file, similarities);
            });
            rowFile.features = rowFile.features.map((file) => {
                return this.addSimilarityInfo(file, similarities);
            });
            rowFile.scenarios = rowFile.scenarios.map((file) => {
                return this.addSimilarityInfo(file, similarities);
            });
            rowFile.states = rowFile.states.map((file) => {
                return this.addSimilarityInfo(file, similarities);
            });
            rowFile.actions = rowFile.actions.map((file) => {
                return this.addSimilarityInfo(file, similarities);
            });
            rowFile.outcomes = rowFile.outcomes.map((file) => {
                return this.addSimilarityInfo(file, similarities);
            });
            return rowFile;
        });
        return rowsFiles;
    }
    addSimilarityInfo(file, similarities) {
        if (similarities[file.id]) {
            file.class = similarities[file.id].class;
            file.similarities = similarities[file.id].similarities;
        }
        return file;
    }
    async readAllTemplates() {
        const templatesNames = ['meta', 'menu', 'footer', 'files', 'features', 'scenarios', 'states', 'actions', 'outcomes'];
        const templates = {};
        await Promise.all(templatesNames
            .map(async (templateName) => {
            const readTemplate = await this.reader.readContentFeatureFile(`${this.folderToReadTemplates}${templateName}.mustache`);
            templates[templateName] = readTemplate;
            return {
                [templateName]: readTemplate
            };
        }));
        return templates;
    }
    prepareReports() {
        this.templatesView = {
            list: this.gherkins,
            date: moment().format('YYYY/MM/DD'),
            time: moment().format('HH:mm:ss')
        };
        this.templatePartials = {
            meta: this.templates.meta,
            footer: this.templates.footer,
            menu: this.templates.menu
        };
    }
    handleError(error) {
        if (error) {
            console.error(error);
            process.exit(1);
            return;
        }
        return;
    }
    writeFilesReport() {
        const reportFilesList = Mustache.render(this.templates.files, this.templatesView, this.templatePartials);
        fs.writeFile(`${this.folderToWriteReport}index.html`, reportFilesList, this.handleError);
    }
    writeFeaturesReport() {
        const reportFeaturesList = Mustache.render(this.templates.features, this.templatesView, this.templatePartials);
        fs.writeFile(`${this.folderToWriteReport}features.html`, reportFeaturesList, this.handleError);
    }
    writeScenariosReport() {
        const reportScenariosList = Mustache.render(this.templates.scenarios, this.templatesView, this.templatePartials);
        fs.writeFile(`${this.folderToWriteReport}scenarios.html`, reportScenariosList, this.handleError);
    }
    writeStatesReport() {
        const reportStatesList = Mustache.render(this.templates.states, this.templatesView, this.templatePartials);
        fs.writeFile(`${this.folderToWriteReport}states.html`, reportStatesList, this.handleError);
    }
    writeActionsReport() {
        const reportActionsList = Mustache.render(this.templates.actions, this.templatesView, this.templatePartials);
        fs.writeFile(`${this.folderToWriteReport}actions.html`, reportActionsList, this.handleError);
    }
    writeOutcomesReport() {
        const reportOutcomesList = Mustache.render(this.templates.outcomes, this.templatesView, this.templatePartials);
        fs.writeFile(`${this.folderToWriteReport}outcomes.html`, reportOutcomesList, this.handleError);
    }
    async reportFeaturesFiles(readError, readFiles) {
        this.handleError(readError);
        this.gherkins = await this.readAllGherkins(readFiles);
        this.templates = await this.readAllTemplates();
        this.prepareReports();
        this.writeFilesReport();
        this.writeFeaturesReport();
        this.writeScenariosReport();
        this.writeStatesReport();
        this.writeActionsReport();
        this.writeOutcomesReport();
    }
}
//# sourceMappingURL=reporter.js.map